void DebugWin::start() {
	emulwin->setpause(true);
	active = true;
	upadr = pc;
	fillall();
	show();
}

void DebugWin::stop() {
	hide();
	active = false;
	emulwin->setpause(false);
}

void DebugWin::reject() {stop();}

#define DASMROW 28
#define DMPSIZE 16

DebugWin::DebugWin(QWidget* par):QDialog(par) {
	QLabel *lab;
	int i,j;
	QHBoxLayout *alay = new QHBoxLayout;
		QVBoxLayout *llay = new QVBoxLayout;
			QGroupBox *regbox = new QGroupBox("Registers");
				rglay = new QGridLayout;
					rglay->setVerticalSpacing(0);
					lab = new QLabel("AF"); rglay->addWidget(lab,0,0); lab = new QLabel; rglay->addWidget(lab,0,1); lab->setAutoFillBackground(true);
					lab = new QLabel("BC"); rglay->addWidget(lab,1,0); lab = new QLabel; rglay->addWidget(lab,1,1); lab->setAutoFillBackground(true);
					lab = new QLabel("DE"); rglay->addWidget(lab,2,0); lab = new QLabel; rglay->addWidget(lab,2,1); lab->setAutoFillBackground(true);
					lab = new QLabel("HL"); rglay->addWidget(lab,3,0); lab = new QLabel; rglay->addWidget(lab,3,1); lab->setAutoFillBackground(true);
					lab = new QLabel("AF'"); rglay->addWidget(lab,4,0); lab = new QLabel; rglay->addWidget(lab,4,1); lab->setAutoFillBackground(true);
					lab = new QLabel("BC'"); rglay->addWidget(lab,5,0); lab = new QLabel; rglay->addWidget(lab,5,1); lab->setAutoFillBackground(true);
					lab = new QLabel("DE'"); rglay->addWidget(lab,6,0); lab = new QLabel; rglay->addWidget(lab,6,1); lab->setAutoFillBackground(true);
					lab = new QLabel("HL'"); rglay->addWidget(lab,7,0); lab = new QLabel; rglay->addWidget(lab,7,1); lab->setAutoFillBackground(true);
					lab = new QLabel("IX"); rglay->addWidget(lab,8,0); lab = new QLabel; rglay->addWidget(lab,8,1); lab->setAutoFillBackground(true);
					lab = new QLabel("IY"); rglay->addWidget(lab,9,0); lab = new QLabel; rglay->addWidget(lab,9,1); lab->setAutoFillBackground(true);
					lab = new QLabel("IR"); rglay->addWidget(lab,10,0); lab = new QLabel; rglay->addWidget(lab,10,1); lab->setAutoFillBackground(true);
					lab = new QLabel("PC"); rglay->addWidget(lab,11,0); lab = new QLabel; rglay->addWidget(lab,11,1); lab->setAutoFillBackground(true);
					lab = new QLabel("SP"); rglay->addWidget(lab,12,0); lab = new QLabel; rglay->addWidget(lab,12,1); lab->setAutoFillBackground(true);
					rglay->setColumnStretch(1,10);
					rglay->setRowStretch(100,10);
					rowincol[0] = 12;
				regbox->setLayout(rglay);
			llay->addWidget(regbox);
			llay->addStretch(10);
		QGroupBox *asmbox = new QGroupBox("Disasm");
			asmlay = new QGridLayout;
				asmlay->setHorizontalSpacing(0);
				asmlay->setVerticalSpacing(0);
				for (i=0;i<DASMROW;i++) {
					lab = new QLabel; asmlay->addWidget(lab,i,0);
						lab->setAutoFillBackground(true);
					lab = new QLabel; asmlay->addWidget(lab,i,1);
						lab->setAutoFillBackground(true);
					lab = new QLabel; asmlay->addWidget(lab,i,2);
						lab->setAutoFillBackground(true);
				}
				rowincol[1] = rowincol[2] = rowincol[3] = DASMROW-1;
				asmlay->setColumnMinimumWidth(0,40);
				asmlay->setColumnMinimumWidth(1,80);
				asmlay->setColumnMinimumWidth(2,130);
			asmbox->setLayout(asmlay);
		QVBoxLayout *rlay = new QVBoxLayout;
			QGroupBox *dmpbox = new QGroupBox("Memory");
				dmplay = new QGridLayout;
				dmplay->setVerticalSpacing(0);
					for (i=0; i<DMPSIZE; i++) {
						lab = new QLabel; dmplay->addWidget(lab,i,0);
						lab->setAutoFillBackground(true);
						for (j=1; j<9; j++) {
							lab = new QLabel;
							lab->setAutoFillBackground(true);
							dmplay->addWidget(lab,i,j);
						}
					}
					for (i=4; i<13; i++) rowincol[i]=DMPSIZE-1;
				dmplay->setColumnMinimumWidth(0,40);
				dmpbox->setLayout(dmplay);
			rlay->addWidget(dmpbox);
			rlay->addStretch(10);
		alay->addLayout(llay);
		alay->addWidget(asmbox);
		alay->addLayout(rlay,10);
	setLayout(alay);

	curcol = 1;
	currow = 0;
	dmpadr = 0;

	QPalette pal;
	pal.setColor(QPalette::ToolTipBase,QColor(128,255,128));
	pal.setColor(QPalette::Highlight,QColor(255,128,128));
	pal.setColor(QPalette::ToolTipText,QColor(255,0,0));
	setPalette(pal);

	setWindowTitle("Xpeccy deBUGer");
	setWindowIcon(QIcon(":/images/bug.png"));
	setFont(QFont("Monospace",9));
	setFixedSize(640,480);
}

QString gethexword(int num) {return QString::number(num+0x10000,16).right(4).toUpper();}
QString gethexbyte(uchar num) {return QString::number(num+0x100,16).right(2).toUpper();}

bool DebugWin::fillall() {
	fillregz();
	filldump();
	return filldasm();
}

void DebugWin::filldump() {
	int i,j;
	adr = dmpadr;
	QLabel *lab;
	for (i=0; i<DMPSIZE; i++) {
		lab = (QLabel*)dmplay->itemAtPosition(i,0)->widget(); lab->setText(gethexword(adr));
		lab->setBackgroundRole((curcol==4 && currow==i)?QPalette::Highlight:QPalette::Window);
		for (j=1; j<9; j++) {
			lab = (QLabel*)dmplay->itemAtPosition(i,j)->widget(); lab->setText(gethexbyte(mem.rd(adr++)));
			lab->setBackgroundRole((curcol==4+j && currow==i)?QPalette::Highlight:QPalette::Window);
		}
	}
}

void DebugWin::fillregz() {
	QLabel *lab;
	lab = (QLabel*)rglay->itemAtPosition(0,1)->widget(); lab->setText(gethexword(getpair(6)));	// af
		lab->setBackgroundRole((curcol==0 && currow==0)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(1,1)->widget(); lab->setText(gethexword(getpair(0)));	// bc
		lab->setBackgroundRole((curcol==0 && currow==1)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(2,1)->widget(); lab->setText(gethexword(getpair(2)));	// de
		lab->setBackgroundRole((curcol==0 && currow==2)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(3,1)->widget(); lab->setText(gethexword(getpair(4)));	// hl
		lab->setBackgroundRole((curcol==0 && currow==3)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(4,1)->widget(); lab->setText(gethexword(getaltpair(6)));	// af'
		lab->setBackgroundRole((curcol==0 && currow==4)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(5,1)->widget(); lab->setText(gethexword(getaltpair(0)));	// bc'
		lab->setBackgroundRole((curcol==0 && currow==5)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(6,1)->widget(); lab->setText(gethexword(getaltpair(2)));	// de'
		lab->setBackgroundRole((curcol==0 && currow==6)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(7,1)->widget(); lab->setText(gethexword(getaltpair(4)));	// hl'
		lab->setBackgroundRole((curcol==0 && currow==7)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(8,1)->widget(); lab->setText(gethexword(getpair(10)));	// ix
		lab->setBackgroundRole((curcol==0 && currow==8)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(9,1)->widget(); lab->setText(gethexword(getpair(11)));	// iy
		lab->setBackgroundRole((curcol==0 && currow==9)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(10,1)->widget(); lab->setText(gethexword((ireg<<8)+rreg));	// ir
		lab->setBackgroundRole((curcol==0 && currow==10)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(11,1)->widget(); lab->setText(gethexword(pc));		// pc
		lab->setBackgroundRole((curcol==0 && currow==11)?QPalette::Highlight:QPalette::Window);
	lab = (QLabel*)rglay->itemAtPosition(12,1)->widget(); lab->setText(gethexword(getpair(8)));	// sp
		lab->setBackgroundRole((curcol==0 && currow==12)?QPalette::Highlight:QPalette::Window);
}

QString DebugWin::getdisasm() {
	uchar cde,mde=0;
	bool prf;
	QString res;
	do {
		prf = false;
		cde = mem.rd(adr++);
		if (mde<3) {
			if (cde==0xdd) {mde=1; prf=true;}
			if (cde==0xfd) {mde=2; prf=true;}
			if (cde==0xcb) {mde+=3; prf=true;}
			if (cde==0xed) {mde=6; prf=true;}
		}
	} while (prf && mde!=4 && mde!=5);
	if ((mde==4) || (mde==5)) {cde=mem.rd(adr++);}
	res = QString(comz[cde][mde].c_str());
	if (res == "") res = QString(comz[cde][0].c_str());
	
	QString nm;
	char bt;
	if (res.indexOf(":4")!=-1) {
		bt = (char)mem.rd(adr++);
		if (bt<0) {nm=QString("-").append(QString::number(512-bt,16).right(2));}
			else {nm=QString("+").append(QString::number(256+bt,16).right(2));}
		res.replace(":4",nm);
	}
	if (res.indexOf(":5")!=-1) {
		bt = (char)mem.rd(adr-2);
		if (bt<0) {nm=QString("-").append(QString::number(512-bt,16).right(2));}
			else {nm=QString("+").append(QString::number(256+bt,16).right(2));}
		res.replace(":5",nm);
	}
	if (res.indexOf(":1")!=-1) res.replace(":1",QString::number(mem.rd(adr++)+0x100,16).right(2).toUpper());
	if (res.indexOf(":2")!=-1) res.replace(":2",gethexword(mem.rd(adr++) + (mem.rd(adr++)<<8)));
	if (res.indexOf(":3")!=-1) res.replace(":3",gethexword(adr + (char)mem.rd(adr++)));
	return res.toUpper();
}

uchar getbpage(ushort ad) {
	uchar res = 0;
	if (ad<0x4000) {res = mem.currom;}
		else {if (ad>0xbfff) res = mem.curram;}
	return res;
}

bool DebugWin::filldasm() {
	adr = upadr;
	ushort oadr;
	uchar i;
	ushort j;
	int idx;
	bool ispc = false;
	QString btez;
	BPoint bp;
	QLabel *lab1,*lab2,*lab3;
	for (i=0; i<DASMROW; i++) {
		oadr = adr;
		bp.page = getbpage(adr);
		bp.adr = adr;
		idx = findbp(bp);
		btez.clear();
		lab1 = (QLabel*)asmlay->itemAtPosition(i,0)->widget();
		lab2 = (QLabel*)asmlay->itemAtPosition(i,1)->widget();
		lab3 = (QLabel*)asmlay->itemAtPosition(i,2)->widget();
		lab1->setBackgroundRole((pc==adr)?QPalette::ToolTipBase:QPalette::Window);
		lab2->setBackgroundRole((pc==adr)?QPalette::ToolTipBase:QPalette::Window);
		lab3->setBackgroundRole((pc==adr)?QPalette::ToolTipBase:QPalette::Window);
		lab1->setForegroundRole((idx==-1)?QPalette::WindowText:QPalette::ToolTipText);
		lab2->setForegroundRole((idx==-1)?QPalette::WindowText:QPalette::ToolTipText);
		lab3->setForegroundRole((idx==-1)?QPalette::WindowText:QPalette::ToolTipText);
		if (adr==pc) ispc=true;
		if (curcol==1 && currow==i) lab1->setBackgroundRole(QPalette::Highlight);
		if (curcol==2 && currow==i) lab2->setBackgroundRole(QPalette::Highlight);
		if (curcol==3 && currow==i) lab3->setBackgroundRole(QPalette::Highlight);
		lab1->setText(gethexword(adr));
		lab3->setText(getdisasm());
		for (j=oadr;j<adr;j++) {btez.append(gethexbyte(mem.rd(j)));}
		lab2->setText(btez.toUpper());
	}
	return ispc;
}

ushort DebugWin::getprevadr(ushort ad) {
	ushort i;
	for (i=16;i>0;i--) {
		adr = ad-i;
		getdisasm();
		if (adr==ad) return (ad-i);
	}
	return (ad-1);
}

int DebugWin::findbp(BPoint bp) {
	int idx=-1,i;
	for (i=0;i<bpoint.size();i++) {
		if ((bpoint[i].page == bp.page) && (bpoint[i].adr == bp.adr)) {idx=i; break;}
	}
	return idx;
}

void executecode();
void DebugWin::keyPressEvent(QKeyEvent* ev) {
	qint32 cod = ev->key();
	QLabel *lab;
	uchar i;
	switch (ev->modifiers()) {
		case Qt::AltModifier:
			switch (cod) {
				case Qt::Key_A: dmpadr = getpair(6); filldump(); break;
				case Qt::Key_B: dmpadr = getpair(0); filldump(); break;
				case Qt::Key_D: dmpadr = getpair(2); filldump(); break;
				case Qt::Key_H: dmpadr = getpair(4); filldump(); break;
				case Qt::Key_X: dmpadr = getpair(10); filldump(); break;
				case Qt::Key_Y: dmpadr = getpair(12); filldump(); break;
				case Qt::Key_S: dmpadr = getpair(8); filldump(); break;
				case Qt::Key_P: dmpadr = pc; filldump(); break;
			}
			break;
		case Qt::NoModifier:
			switch (cod) {
				case Qt::Key_Escape: stop(); break;
				case Qt::Key_Space:
					if (!ev->isAutoRepeat() && curcol>0 && curcol<4) {
						QLabel *lab = (QLabel*)asmlay->itemAtPosition(currow,0)->widget();
						BPoint bp;
						bp.adr = lab->text().toInt(NULL,16);
						bp.page = getbpage(bp.adr);
						int idx = findbp(bp);
						if (idx==-1) bpoint.append(bp); else bpoint.removeAt(idx);
						filldasm();
					}
					break;
				case Qt::Key_F7:
					executecode();
					if (!fillall()) {upadr = pc; filldasm();}
					break;
				case Qt::Key_Left:
					if (curcol>0) {
						curcol--;
						if (currow >= rowincol[curcol]) currow = rowincol[curcol];
					}
					fillall();
					break;
				case Qt::Key_Right:
					if (curcol<12) {
						curcol++;
						if (currow >= rowincol[curcol]) currow = rowincol[curcol];
					}
					fillall();
					break;
				case Qt::Key_Down:
					if (currow < rowincol[curcol]) {
						currow++;
					} else {
						if (curcol>0 && curcol<4) {
							lab = (QLabel*)asmlay->itemAtPosition(1,0)->widget();
							upadr = lab->text().toInt(NULL,16);
						}
						if (curcol>3 && curcol<12) dmpadr += 8;
					}
					fillall();
					break;
				case Qt::Key_Up:
					if (currow > 0) {
						currow--;
					} else {
						if (curcol>0 && curcol<4) upadr = getprevadr(upadr);
						if (curcol>3 && curcol<12) dmpadr -= 8;
					}
					fillall();
					break;
				case Qt::Key_PageDown:
					if (curcol>0 && curcol<4) {
						lab = (QLabel*)asmlay->itemAtPosition(DASMROW-1,0)->widget();
						upadr = lab->text().toInt(NULL,16);
						filldasm();
					}
					if (curcol>3 && curcol<12) {
						dmpadr += (DMPSIZE << 3);
						filldump();
					}
					break;
				case Qt::Key_PageUp:
					if (curcol>0 && curcol<4) {
						for (i=0; i<DASMROW-2; i++) upadr=getprevadr(upadr);
						filldasm();
					}
					if (curcol>3 && curcol<12) {
						dmpadr -= (DMPSIZE << 3);
						filldump();
					}
					break;
				case Qt::Key_Home:
					upadr = pc;
					filldasm();
					break;
			}
			break;
	}
}