DevelWin::DevelWin(QWidget *parent=0):QDialog(parent) {
	QVBoxLayout *mainlay = new QVBoxLayout;
	QHBoxLayout *uplay = new QHBoxLayout;
	newprj = new QToolButton; newprj->setIcon(QIcon(":/images/fileopen.png"));
	savbut = new QToolButton; savbut->setIcon(QIcon(":/images/save_all.png"));
	compbut = new QToolButton; compbut->setIcon(QIcon(":/images/run.png"));
	uplay->addWidget(newprj);
	uplay->addWidget(savbut);
	uplay->addSpacing(10);
	uplay->addWidget(compbut);
	uplay->addStretch();
	mainlay->addLayout(uplay);
	QHBoxLayout *midlay = new QHBoxLayout;
	QVBoxLayout *treelay = new QVBoxLayout;
	tree = new MyListView; tree->setEditTriggers(QAbstractItemView::NoEditTriggers);
	treelay->addWidget(tree);
	QHBoxLayout *trdnlay = new QHBoxLayout;
	newfile = new QToolButton; newfile->setIcon(QIcon(":/images/add.png"));
	addfile = new QToolButton; addfile->setIcon(QIcon(":/images/fileopen.png"));
	delfile = new QToolButton; delfile->setIcon(QIcon(":/images/remove.png"));
	inpfname = new QLineEdit;
	trdnlay->addWidget(inpfname);
	trdnlay->addWidget(newfile);
	trdnlay->addWidget(addfile);
	trdnlay->addWidget(delfile);
	treelay->addLayout(trdnlay);
	editor = new QTextEdit; editor->setLineWrapMode(QTextEdit::NoWrap); hlight = new MHighlight(editor);
	midlay->addLayout(treelay);
	midlay->addWidget(editor); midlay->setStretchFactor(editor,100);
	mainlay->addLayout(midlay);
	cons = new QTextEdit; cons->setReadOnly(true); cons->setFixedHeight(150);
	mainlay->addWidget(cons);

	filetree = new QTreeView;

	setLayout(mainlay);
	setWindowTitle("X-peccy developer tool");
	resize(800,600);

	QObject::connect(newprj,SIGNAL(released()),this,SLOT(makenewproject()));
	QObject::connect(savbut,SIGNAL(released()),this,SLOT(saveproject()));
	QObject::connect(newfile,SIGNAL(released()),this,SLOT(addnewfiletoproject()));
	QObject::connect(addfile,SIGNAL(released()),this,SLOT(addfiletoproject()));
	QObject::connect(delfile,SIGNAL(released()),this,SLOT(delfilefromproject()));
	QObject::connect(tree,SIGNAL(rowchanged(int,int)),this,SLOT(recalltext(int,int)));
	QObject::connect(filetree,SIGNAL(doubleClicked(QModelIndex)),this,SLOT(trueaddfiletoproject(QModelIndex)));
	QObject::connect(compbut,SIGNAL(released()),this,SLOT(compile()));
}

void DevelWin::addnewfiletoproject() {
	if (basepath=="") return;
	QString name = inpfname->text();
	if (name=="") return;
	if (!name.endsWith(".asm")) name.append(".asm");
	inpfname->clear();
	bool ok=false; int i; for (i=0;i<project.size();i++) {ok = ok || (project[i].name == name);}
	if (ok) {
		QMessageBox mbox(this); mbox.setText("This file is already in project"); mbox.exec();
	} else {
		ProjectFile newfile;
		newfile.name = name;
		newfile.changed=false;
		QFile file(QString(basepath).append("/").append(name));
		if (file.exists()) {file.open(QFile::ReadOnly); newfile.text = QDialog::trUtf8(file.readAll()); file.close();}
		else {file.open(QFile::WriteOnly); file.close(); newfile.text = "";}
		project.append(newfile);
		saveprojectfile();
		updateprojectlist();
	}
}

void DevelWin::compile() {
	if (project.size()==0) return;
	QString output;
	saveproject();
	if (sjapath=="") {output = "No SJASM path specified.";}
	else {
		QProcess proc; QStringList parm; parm<<QString(basepath).append("/").append("main.asm");
		proc.setWorkingDirectory(basepath);
		proc.start(QString(sjapath.c_str()),parm);
		do {} while(proc.waitForFinished(60000));	// 1 minute for finishing compilation
		output = QDialog::trUtf8(proc.readAll());
	}
	cons->setText(output);
}

void DevelWin::trueaddfiletoproject(QModelIndex idx) {
	QDirModel *model = (QDirModel*)(filetree->model());
	QString fullpath = model->filePath(idx);
	QString name = QString(fullpath).remove(0,basepath.size()+1);
	bool ok=false;
	int i; for(i=0;i<project.size();i++) {ok = ok || (project[i].name==name);}
	if (ok) {
		QMessageBox mbox(filetree); mbox.setText("This file is already in project"); mbox.exec();
	} else {
		ProjectFile newfile;
		newfile.name = fullpath.remove(0,basepath.size()+1);
		newfile.changed=false;
		QFile file(fullpath); file.open(QFile::ReadOnly); newfile.text = QDialog::trUtf8(file.readAll()); file.close();
		project.append(newfile); saveprojectfile();
		filetree->hide();
	}
}

void DevelWin::addfiletoproject() {
	if (basepath=="") return;
	QDirModel *model = new QDirModel; model->setNameFilters(QStringList()<<"*.asm"); model->sort(QDir::Name);
	filetree->setModel(model);
	filetree->setRootIndex(model->index(QString(basepath).append("/")));
	filetree->show();
}

void DevelWin::delfilefromproject() {
	if (project.size()==0) return;
	int idx = tree->firstselectedrow();
	if (idx==-1) return;
	if (project[idx].name=="main.asm") return;
	project.removeAt(idx);
	if (currentfile>(project.size()-1)) currentfile--;
	editor->setText(project[currentfile].text);
	saveprojectfile();
	updateprojectlist();
}

void DevelWin::saveprojectfile() {
	if (project.size()==0) return;
	QFile file(QString(basepath).append("/").append("project.lst"));
	file.open(QFile::WriteOnly);
	int i; for(i=0;i<project.size();i++) {file.write(QString(project[i].name).append("\r\n").toUtf8());}
	file.close();
}

void DevelWin::saveproject() {
	if (project.size()==0) return;
	QFile file;
	storecurrentfile();
	int i; for (i=0;i<project.size();i++) {
		file.setFileName(QString(basepath).append("/").append(project[i].name));
		file.open(QFile::WriteOnly); file.write(project[i].text.toUtf8()); file.close();
		project[i].changed=false;
	}
	updateprojectlist();
}

void DevelWin::recalltext(int newidx,int) {
	if (newidx<0 || newidx>=project.size()) return;
	storecurrentfile();
	updateprojectlist();
	currentfile = newidx; editor->setText(project[currentfile].text);
}

void DevelWin::updateprojectlist() {
	QStringListModel *model = new QStringListModel;
	QStringList list; QString item; int i;
	for (i=0;i<project.size();i++) {
		item=project[i].name;
		if (project[i].changed) item.prepend("* ");
		list<<item;
	}
	model->setStringList(list);
	tree->setModel(model);
}

void DevelWin::getproject() {
	QFile file(QString(basepath).append("/project.lst"));
	QFile itemfile; QString line;
	ProjectFile newfile;
	project.clear();
	file.open(QFile::ReadOnly);
	while(!file.atEnd()) {
		line = QDialog::trUtf8(file.readLine()).remove(QRegExp("[\r\n]"));
		if (line!="") {
			newfile.name = line;
			itemfile.setFileName(QString(basepath).append("/").append(newfile.name));
			if (itemfile.exists()) {
				itemfile.open(QFile::ReadOnly); newfile.text = QDialog::trUtf8(itemfile.readAll()); itemfile.close();
			} else {newfile.text = "";itemfile.open(QFile::WriteOnly);itemfile.close();}
			newfile.changed = false;
			project.append(newfile);
		}
	}
	file.close();
	currentfile = 0;
	if (project.size()!=0) editor->setText(project[0].text);
	updateprojectlist();
}

void DevelWin::makenewproject() {
	QString folder = QFileDialog::getExistingDirectory(this,"Choose project directory",QDir::homePath());
	if (folder=="") return;
	basepath = folder;
	QFile prjfile(QString(basepath).append("/project.lst"));
	if (!prjfile.exists()) {prjfile.open(QFile::WriteOnly); prjfile.write(QByteArray("main.asm\r\n")); prjfile.close();}
	getproject();
}

void DevelWin::storecurrentfile() {
	project[currentfile].changed = project[currentfile].changed || (project[currentfile].text != editor->toPlainText());
	project[currentfile].text = editor->toPlainText();
}

void DevelWin::reject() {
	if (project.size()!=0) storecurrentfile();
	hide();
}

void MHighlight::highlightBlock(const QString &text) {
	int sidx,eidx;
	QTextCharFormat txthi; txthi.setForeground(Qt::red);
	sidx = text.indexOf("\"");
	while (sidx>=0) {
		eidx = text.indexOf("\"",sidx+1); if (eidx==-1) eidx=text.size();
		setFormat(sidx+1,eidx-sidx-1,txthi);
		sidx=text.indexOf("\"",eidx+1);
	}
}