EmulWindow::EmulWindow(QWidget *parent=0):QGLWidget(parent) {
	setWindowIcon(QIcon(":/images/logo.svg"));
	setWindowTitle("Xpeccy 0.2");
	setFixedSize(xscrsize*(doublesize?2:1),yscrsize*(doublesize?2:1));
	QGLFormat glformat;glformat.setDoubleBuffer(true); glformat.setAlpha(true); setFormat(glformat);

	setMouseTracking(true);
	paused = false;
	mgrab = false;
	menu = new QMenu("User Menu",this);
	makemenu();
}

static int int_log2(int val) {
	int log = 0;
	while ((val >>= 1) != 0) log++;
	return log;
}

void EmulWindow::resizeGL(int, int) {
	glMatrixMode(GL_PROJECTION);
	glDeleteTextures(1, &texture);
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	// No borders
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_LINEAR);
	// TODO: Make an option: GL_LINEAR or GL_NEAREST
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	int texsize = 2 << int_log2(xscrsize > yscrsize ? xscrsize : yscrsize);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, texsize, texsize, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);

	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT);
	swapBuffers();
	glClear(GL_COLOR_BUFFER_BIT);
	glShadeModel(GL_FLAT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	GLfloat tex_width = ((GLfloat)(xscrsize)/(GLfloat)texsize);
	GLfloat tex_height = ((GLfloat)(yscrsize)/(GLfloat)texsize);

	if (glIsList(displaylist)) glDeleteLists(displaylist, 1);
	displaylist = glGenLists(1);
	glNewList(displaylist, GL_COMPILE);
	glBindTexture(GL_TEXTURE_2D, texture);
	glBegin(GL_QUADS);
	// lower left
	glTexCoord2f(0, tex_height); glVertex2f(-1.0f, -1.0f);
	// lower right
	glTexCoord2f(tex_width, tex_height); glVertex2f(1.0f, -1.0f);
	// upper right
	glTexCoord2f(tex_width, 0); glVertex2f(1.0f, 1.0f);
	// upper left
	glTexCoord2f(0,0); glVertex2f(-1.0f, 1.0f);
	glEnd();
	glEndList();
}

void EmulWindow::setpause(bool p) {
	paused = p;
	if (p && mgrab) releaseMouse();
	if (!p && mgrab) grabMouse(QCursor(Qt::BlankCursor));
}

void EmulWindow::closeEvent(QCloseEvent* ev) {
	uchar i; for (i=0;i<4;i++) {
		if (!flop[i].savecha()) {ev->ignore(); return;}
	}
	ev->accept();
}

void EmulWindow::paintmain() {
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, xscrsize, yscrsize, GL_RGB,GL_UNSIGNED_BYTE, screen);
	glCallList(displaylist);
	if (paused) renderText(0,10,"PAUSE");
	if (tape.play) renderText(0,20,"TAPE PLAY");
	if (tape.rec) renderText(0,20,"TAPE REC");
	swapBuffers();
	if (paused) return;
	
	if (sshotbckcount!=0) {
		if (sshotbcktime==0) {needsavescreenshot=true; sshotbckcount--; sshotbcktime=sshotrpttime; if (sshotbckcount==0) qDebug()<<"stop combo shots";}
			else {sshotbcktime--;}
	}
	if (needsavescreenshot) {
		std::string fnam = screenshotdir + "sshot" + int2str(screenshotnum) + "." + screenshotformat;
//		filename.append("sshot").append(QString::number(screenshotnum+10000).right(4)).append(".").append(screenshotformat);
		if (screenshotformat == "scr") {
			std::ofstream file(fnam.c_str(),std::ios::binary);
			file.write((char*)&mem.ram[curscrn?7:5][0],0x1b00);
		} else {
			QImage *img = new QImage(&screen[0],xscrsize,yscrsize,QImage::Format_RGB888);
			if (img==NULL) {
				printf("NULL image\n");
			} else {
				img->save(QString(fnam.c_str()),screenshotformat.c_str());
			}
		}
		needsavescreenshot=false;
		screenshotnum++;
	}

}

void EmulWindow::emulframe() {
	if (!fullscrn && isFullScreen()) showNormal();
	if (fullscrn && !isFullScreen()) showFullScreen();
	if (!fullscrn && (size().width()!=xscrsize*(doublesize?2:1) || size().height()!=yscrsize*(doublesize?2:1))) {
		setFixedSize(xscrsize*(doublesize?2:1),yscrsize*(doublesize?2:1));
	}
	glViewport(0,0,size().width(),size().height());
	if (soundenabled && (inactivesnd || isActiveWindow()) && sndbufpos>1) {cursoundsys->play();}
	if (!isActiveWindow()) {
		releasekeys();		// в неактивном окне все кнопки отпущены
		if (mouseGrabber()==this) releaseMouse();
	} else {
		if (mgrab && mouseGrabber()==NULL) grabMouse(QCursor(Qt::BlankCursor));
		if (!mgrab && mouseGrabber()==this) releaseMouse();
	}
	intupt = false; hcount=0; vcount=0;
	scrptr = &screen[0];	// dlinepos
	sndbufpos=0;
	if (!paused) execute();
	paintmain();
}

int findkey(qint32 code) {
	std::vector<qint32> res(4,0);
	int idx = 0;
	while (keyz[idx][0]!=0) {
		if (keyz[idx][0]==code) return idx;
		idx++;
	}
	return -1;
}

void savesettings();
void EmulWindow::keyPressEvent(QKeyEvent* event) {
	if (event->isAutoRepeat()) return;
	quint32 code=event->key();
	if (code==Qt::Key_Escape) {
		if (!dbg->active) {dbg->start();} else {dbg->stop();}
	};
	if (event->modifiers()==Qt::AltModifier) {
		switch (code) {
			case Qt::Key_Return:
				if (isFullScreen()) {showNormal(); fullscrn=false;}
					else {showFullScreen(); fullscrn=true;}
				savesettings();
				break;
			case Qt::Key_1: doublesize=false; fullscrn=false; showNormal(); savesettings(); break;
			case Qt::Key_2: doublesize=true; fullscrn=false; showNormal(); savesettings(); break;
			case Qt::Key_F7: sshotbckcount=sshotrptcount; sshotbcktime=0; break;	// Alt+F7 запустить автоповтор скриншотинга
		}
		return;
	}
	if (code==Qt::Key_Menu) {menu->popup(pos() + QPoint(20,20));}
	if (!paused && !dbg->active) {
		switch (code) {
			case Qt::Key_F1: paused=true; emit(wannasetup()); break;	// F1 (67)
			case Qt::Key_F2: paused=true; saveonf2(); paused=false; break;		// F2
			case Qt::Key_F3: paused=true; opensomewhat(); paused=false; break;		// F3
			case Qt::Key_F4: if (tape.rec) {tape.rec=false; tape.play=false;}
					else {if(tape.block<tape.data.size()) {tape.ticks=0;tape.pos=0;tape.play=!tape.play;}
						else {tape.play=false;}
				} break;	// F4: play tape / stop tape
			case Qt::Key_F5: if (tape.play) {tape.play=false; tape.rec=false;}
					else {	if (tape.rec) {tape.rec=false; printf("stop tape\n");} else {
							printf("start tape\n");
							tape.rec=true; tape.toutold=tape.outsig;
							tape.blockptr.data.clear();
							tape.blockptr.data.push_back(0);
						}
					} break;	// F5: rec tape / stop tape
			case Qt::Key_F6: emit wannadevelop(); break;
			case Qt::Key_F7: needsavescreenshot=(sshotbckcount==0); break;	// F8 в конце кадра щелкнуть скриншот (если не автоповтор)
			case Qt::Key_F9: flop[0].savecha();
					flop[1].savecha();
					flop[2].savecha();
					flop[3].savecha();
					break;	// F9
		}
	}

	if (!paused) {
		int idx = findkey(code);
//		printf("%.8X\n",code);
		if (event->modifiers() & Qt::KeypadModifier) {
			if (code==Qt::Key_0 || code==Qt::Key_Insert) gfxmode = !gfxmode;
			if (code==Qt::Key_1 || code==Qt::Key_End) traceout = !traceout;
			if (code==Qt::Key_3 || code==Qt::Key_PageDown) curscrn = !curscrn;
		} else {
			if (code==Qt::Key_F12) reset();
			if (idx!=-1) {
				keyboardmap[keyz[idx][1]] = keyboardmap[keyz[idx][1]] & (~keyz[idx][2]);
				keyboardmap[keyz[idx][3]] = keyboardmap[keyz[idx][3]] & (~keyz[idx][4]);
			}
		}
	}
}

void EmulWindow::keyReleaseEvent(QKeyEvent* event) {
	if (event->isAutoRepeat() || paused) return;
	int idx = findkey(event->key());
	keyboardmap[keyz[idx][1]] = keyboardmap[keyz[idx][1]] | keyz[idx][2];
	keyboardmap[keyz[idx][3]] = keyboardmap[keyz[idx][3]] | keyz[idx][4];
}

void EmulWindow::makemenu() {
	menu->clear();
	uint i;
	QAction *act;
	for (i=0; i<usermenu.size(); i++) {
		act = menu->addAction(QString(usermenu[i].name.c_str()));
		act->setToolTip(QString(usermenu[i].path.c_str()));
	}
	QObject::connect(menu,SIGNAL(triggered(QAction*)),this,SLOT(actmenu(QAction*)));
}

void EmulWindow::actmenu(QAction* act) {
	loadsomefile(act->toolTip().toUtf8().data()); setFocus();
	if (mgrab) setCursor(QCursor(Qt::BlankCursor));
}

void EmulWindow::mousePressEvent(QMouseEvent *ev) {
	if (ev->button()==Qt::MidButton) {mgrab = !mgrab; if (!mgrab) setCursor(QCursor());}
	if (!mgrab) return;
	if (ev->button()==Qt::LeftButton) mouse.buttons &= 0xfe;
	if (ev->button()==Qt::RightButton) mouse.buttons &= 0xfd;
}

void EmulWindow::mouseReleaseEvent(QMouseEvent *ev) {
	if (!mgrab) return;
	if (ev->button()==Qt::LeftButton) mouse.buttons |= 0x01;
	if (ev->button()==Qt::RightButton) mouse.buttons |= 0x02;
}

void EmulWindow::mouseMoveEvent(QMouseEvent *ev) {
	if (!mgrab) return;
	mouse.xpos = (ev->pos().x() & 0xff);
	mouse.ypos = 255 - ((ev->pos().y()) & 0xff);
//	QTest::mouseMove(this,QPoint(xps,yps));
}